class VectorIndex:
  @property
  def files(self):
    return []

  def save(self):
    pass

  def load(self):
    pass

  def build(self):
    pass

  def search(self):
    pass

  def search_index(self):
    pass

  def add(self, vector):
    pass

  def add_bulk(self, vectors):
    pass

  def set_bulk(self, indices, vectors):
    pass
